let compose f g x = (f (g x))

let rec ppTerm t = match t with
  Var x -> x |
  Fn (name,[]) -> name |
  Fn (name,args) -> name ^ "(" ^ (String.concat "," (map ppTerm args)) ^ ")"


let ppAtom a = match a with
  R (name,[]) -> name |
  R (name,args) -> name ^ "(" ^ (String.concat "," (map ppTerm args)) ^ ")"

let rec ppFormula f = match f with
 Not n -> "~("^(ppFormula n)^")" |
 Atom a -> ppAtom a |
 Forall (var, sf) -> "ALL " ^ var ^ ". (" ^ (ppFormula sf) ^ ")" |
 Exists (var, sf) -> "EX " ^ var ^ ". (" ^ (ppFormula sf) ^ ")" |
 Iff (lhs,rhs) -> "(" ^ (ppFormula lhs) ^ ") <=> (" ^ (ppFormula rhs) ^ ")" |
 Imp (lhs,rhs) -> "(" ^ (ppFormula lhs) ^ ") => (" ^ (ppFormula rhs) ^ ")" |
 And (lhs,rhs) -> "(" ^ (ppFormula lhs) ^ ") /\ (" ^ (ppFormula rhs) ^ ")" |
 Or (lhs,rhs) -> "(" ^ (ppFormula lhs) ^ ") \/ (" ^ (ppFormula rhs) ^ ")" |
 True -> "True" |
 False -> "False"

let ppTLL tll = "[" ^ (String.concat ", " (map (fun ss -> "[" ^ (String.concat "," (map ppFormula ss)) ^ "]") tll)) ^ "]"


let rec escape s = if String.contains s '_'
 then (String.set s (String.index s '_') '-' ; escape s)
 else s

let norm f = simpcnf (skolemize (pnf (Not(generalize f)))) ;;

let ourRes cnfl = []

let rec formatTerm t = match t with
        Var x -> x^":S" |
        Fn (name,args) -> match args with
                              [] -> (escape name) |
                              _  -> (escape name)
                                   ^"("^(String.concat "," (map formatTerm args))^")"

let rec formatAtom a = match a with
        Not  x -> formatAtom x |
        Atom (R (name,args)) -> (match args with
                                  [] -> name |
                                  _  -> name^"(" ^(String.concat "," (map formatTerm args)) ^")") |
        Forall _ -> failwith "Unhandled Forall"

let isNot a = match a with Not _ -> true | _ -> false

let fixEmptyMaps s = match s with
  "" -> "(empty).AtomMagma" |
  _ -> s

let formatClause cl = "(" ^ (fixEmptyMaps (String.concat "," (map formatAtom (filter isNot cl))))
    ^" -> "^
   (fixEmptyMaps (String.concat "," (map formatAtom (filter (compose not isNot) cl)))) ^ ")"

let formatFormula f = "red ?((\n  " ^ (String.concat ",\n  " (map formatClause f)) ^ "\n)) .\n\nq .\n";;

let nonCrapEq (n1,l1) (n2,l2) = ((String.compare n1 n2) == 0) && (l1 = l2);;

let rec uniq l = match l with
 [] -> [] |
 (x::xs) -> (x::uniq (filter (fun y -> not (nonCrapEq x y)) xs));;

let replicate n x = match n with
 0 -> [] |
 _ -> (x::replicate (n-1) x)

let rec getPred a = match a with
       Atom (R (name,args)) -> (escape name,length args) |
       Not a -> getPred a 

let predNumber = ref 0

let formatPreds f = let preds = uniq (List.concat (map (map getPred) f)) in
 String.concat "\n" (map (fun (n,l) -> "  op " ^ n ^ " : " ^
 (String.concat " " (replicate l "S")) ^ " -> Atom [metadata \""^ 
 (string_of_int (incr predNumber; !predNumber))
 ^"\"] .") preds)

let rec getFuncs t = match t with
 Var _ -> [] |
 Fn (name,args) -> ((escape name,length args) :: (List.concat (map getFuncs args)))

let rec getFuncsP a = match a with
 Not x -> getFuncsP x |
 Atom (R (_,args)) -> List.concat (map getFuncs args)

let funcNumber = ref 0

let formatFuncs f = let funcs = uniq (List.concat (List.concat (map (map (getFuncsP)) f))) in
 String.concat "\n" (map (fun (n,l) -> "  op " ^ n ^ " : " ^
 (String.concat " " (replicate l "S")) ^ " -> S [metadata \""^ 
 (string_of_int (incr funcNumber; !funcNumber))
 ^"\"] .") funcs)

let formatDecls f = (formatPreds f) ^ "\n" ^ (formatFuncs f)

let modname name = String.uppercase (escape name)

let usefulTime f x = 
  let start_time = Sys.time() in
  let result = f x in
  let finish_time = Sys.time() in
  (finish_time -. start_time,result);;

let formatMain name f = let cnf = norm f 
and (runtime,harrison) =  usefulTime resolution f
in 
"load ../../rtp\n\nfmod " ^ (modname name) ^ 
" is\npr PRED-BASE .\n\nsort S .\n" ^ formatDecls cnf ^ "\nendfm\n\n" ^
"fmod TEST is\n  inc RTP .\n  inc " ^ (modname name) ^ " .\n\n" ^
"  eq USER-MODULE-NAME = \'" ^ (modname name) ^ " .\nendfm\n\n" ^
(formatFormula cnf) ^ 
"--- Input: " ^ (ppFormula f) ^
"\n--- Generalize: " ^ (ppFormula (generalize f)) ^
"\n--- After Negation: " ^ (ppFormula (Not(generalize f))) ^
"\n--- After Prenex: " ^ (ppFormula (pnf(Not(generalize f)))) ^
"\n--- After Skolemization: " ^ (ppFormula (skolemize (pnf(Not (generalize f))))) ^
"\n--- After simpcnf: " ^ (ppTLL (simpcnf (skolemize (pnf(Not (generalize f)))))) ^
"\n--- Harrison resolution returned [" ^ (String.concat ", " (map string_of_bool (harrison))) ^ "] taking " ^ string_of_float runtime ^
"\n--- Our Harrison resolution solved: " ^ string_of_bool (resloop([],norm f)) ^"\n"

let trans name f = let chan = open_out (name ^".maude") in 
 (printf "Translating: %s\n" name; output_string chan (formatMain name f); flush; close_out chan)

let davis_putnam = <<exists x. exists y. forall z.
        (F(x,y) ==> (F(y,z) /\ F(z,z))) /\
        ((F(x,y) /\ G(x,y)) ==> (G(x,z) /\ G(z,z)))>>;;

trans "davis_putnam" davis_putnam;;

let los =
 <<(forall x y z. P(x,y) ==> P(y,z) ==> P(x,z)) /\
   (forall x y z. Q(x,y) ==> Q(y,z) ==> Q(x,z)) /\
   (forall x y. Q(x,y) ==> Q(y,x)) /\
   (forall x y. P(x,y) \/ Q(x,y))
   ==> (forall x y. P(x,y)) \/ (forall x y. Q(x,y))>>;;

(* Quite slow *)
(* trans "los" los *)

let p1 = <<p ==> q <=> ~q ==> ~p>>;;

trans "p_one" p1

let p2 = <<~ ~p <=> p>>;;

trans "p_two" p2

let p3 = <<~(p ==> q) ==> q ==> p>>;;

trans "p_three" p3

let p4 = <<~p ==> q <=> ~q ==> p>>;;

trans "p_four" p4

let p5 = <<(p \/ q ==> p \/ r) ==> p \/ (q ==> r)>>;;

trans "p_five" p5

let p6 = <<p \/ ~p>>;;

trans "p_six" p6;;

let p7 = <<p \/ ~ ~ ~p>>;;

trans "p_seven" p7

let p8 = <<((p ==> q) ==> p) ==> p>>;;

trans "p_eight" p8

let p9 = <<(p \/ q) /\ (~p \/ q) /\ (p \/ ~q) ==> ~(~q \/ ~q)>>;;

trans "p_nine" p9

let p10 = <<(q ==> r) /\ (r ==> p /\ q) /\ (p ==> q /\ r) ==> (p <=> q)>>;;

trans "p_ten" p10

let p11 = <<p <=> p>>;;

trans "p_eleven" p11

let p12 = <<((p <=> q) <=> r) <=> (p <=> (q <=> r))>>;;

trans "p_twelve" p12

let p13 = <<p \/ q /\ r <=> (p \/ q) /\ (p \/ r)>>;;

trans "p_thirteen" p13

let p14 = <<(p <=> q) <=> (q \/ ~p) /\ (~q \/ p)>>;;

trans "p_fourteen" p14

let p15 = <<p ==> q <=> ~p \/ q>>;;

trans "p_fifteen" p15

let p16 = <<(p ==> q) \/ (q ==> p)>>;;

trans "p_sixteen" p16

let p17 = <<p /\ (q ==> r) ==> s <=> (~p \/ q \/ s) /\ (~p \/ ~r \/ s)>>;;

trans "p_seventeen" p17

(* ------------------------------------------------------------------------- *)
(* Monadic Predicate Logic.                                                  *)
(* ------------------------------------------------------------------------- *)

let p18 = <<exists y. forall x. P(y) ==> P(x)>>;;

trans "p_eightteen" p18

let p19 = <<exists x. forall y z. (P(y) ==> Q(z)) ==> P(x) ==> Q(x)>>;;

trans "p_nineteen" p19

let p20 = <<(forall x y. exists z. forall w. P(x) /\ Q(y) ==> R(z) /\ U(w))
   ==> (exists x y. P(x) /\ Q(y)) ==> (exists z. R(z))>>;;

trans "p_twenty" p20

let p21 = <<(exists x. P ==> Q(x)) /\ (exists x. Q(x) ==> P)
   ==> (exists x. P <=> Q(x))>>;;

trans "p_twentyone" p21

let p22 = <<(forall x. P <=> Q(x)) ==> (P <=> (forall x. Q(x)))>>;;

trans "p_twentytwo" p22

let p23 = <<(forall x. P \/ Q(x)) <=> P \/ (forall x. Q(x))>>;;

trans "p_twentythree" p23

let p24 = <<~(exists x. U(x) /\ Q(x)) /\
   (forall x. P(x) ==> Q(x) \/ R(x)) /\
   ~(exists x. P(x) ==> (exists x. Q(x))) /\
   (forall x. Q(x) /\ R(x) ==> U(x)) ==>
   (exists x. P(x) /\ R(x))>>;;

trans "p_twentyfour" p24

let p25 = 
 <<(exists x. P(x)) /\
   (forall x. U(x) ==> ~G(x) /\ R(x)) /\
   (forall x. P(x) ==> G(x) /\ U(x)) /\
   ((forall x. P(x) ==> Q(x)) \/ (exists x. Q(x) /\ P(x))) ==>
   (exists x. Q(x) /\ P(x))>>;;

trans "p_twentyfive" p25

let p26 =
 <<((exists x. P(x)) <=> (exists x. Q(x))) /\
   (forall x y. P(x) /\ Q(y) ==> (R(x) <=> U(y))) ==>
   ((forall x. P(x) ==> R(x)) <=> (forall x. Q(x) ==> U(x)))>>;;

trans "p_twentysix" p26

let p27 = 
 <<(exists x. P(x) /\ ~Q(x)) /\
   (forall x. P(x) ==> R(x)) /\
   (forall x. U(x) /\ V(x) ==> P(x)) /\
   (exists x. R(x) /\ ~Q(x)) ==>
   (forall x. U(x) ==> ~R(x)) ==>
   (forall x. U(x) ==> ~V(x))>>;;

trans "p_twentyseven" p27

let p28 =
 <<(forall x. P(x) ==> (forall x. Q(x))) /\
   ((forall x. Q(x) \/ R(x)) ==> (exists x. Q(x) /\ R(x))) /\
   ((exists x. R(x)) ==> (forall x. L(x) ==> M(x))) ==>
   (forall x. P(x) /\ L(x) ==> M(x))>>;;

trans "p_twentyeight" p28

let p29 =
 <<(exists x. P(x)) /\ (exists x. G(x)) ==>
   ((forall x. P(x) ==> H(x)) /\ (forall x. G(x) ==> J(x)) <=>
    (forall x y. P(x) /\ G(y) ==> H(x) /\ J(y)))>>;;

trans "p_twentynine" p29

let p30 =
 <<(forall x. P(x) \/ G(x) ==> ~H(x)) /\
   (forall x. (G(x) ==> ~U(x)) ==> P(x) /\ H(x)) ==>
   (forall x. U(x))>>;;

trans "p_thirty" p30

let p31 =
 <<~(exists x. P(x) /\ (G(x) \/ H(x))) /\ (exists x. Q(x) /\ P(x)) /\
   (forall x. ~H(x) ==> J(x)) ==>
   (exists x. Q(x) /\ J(x))>>;;

trans "p_thirtyone" p31

let p32 =
 <<(forall x. P(x) /\ (G(x) \/ H(x)) ==> Q(x)) /\
   (forall x. Q(x) /\ H(x) ==> J(x)) /\
   (forall x. R(x) ==> H(x)) ==>
   (forall x. P(x) /\ R(x) ==> J(x))>>;;

trans "p_thirtytwo" p32

let p33 =
 <<(forall x. P(a) /\ (P(x) ==> P(b)) ==> P(c)) <=>
   (forall x. P(a) ==> P(x) \/ P(c)) /\ (P(a) ==> P(b) ==> P(c))>>;;

trans "p_thirtythree" p33;;

let p34 =
 <<((exists x. forall y. P(x) <=> P(y)) <=>
    ((exists x. Q(x)) <=> (forall y. Q(y)))) <=>
   ((exists x. forall y. Q(x) <=> Q(y)) <=>
    ((exists x. P(x)) <=> (forall y. P(y))))>>;; 

(* trans "p_thirtyfour" p34 *)

let p35 =
 <<exists x y. P(x,y) ==> (forall x y. P(x,y))>>;;

trans "p_thirtyfive" p35

(* ------------------------------------------------------------------------- *)
(*  Full predicate logic (without Identity and Functions)                    *)
(* ------------------------------------------------------------------------- *)

let p36 =
 <<(forall x. exists y. P(x,y)) /\
   (forall x. exists y. G(x,y)) /\
   (forall x y. P(x,y) \/ G(x,y)
   ==> (forall z. P(y,z) \/ G(y,z) ==> H(x,z)))
       ==> (forall x. exists y. H(x,y))>>;;

trans "p_thirtysix" p36

let p37 =
 <<(forall z.
     exists w. forall x. exists y. (P(x,z) ==> P(y,w)) /\ P(y,z) /\
     (P(y,w) ==> (exists u. Q(u,w)))) /\
   (forall x z. ~P(x,z) ==> (exists y. Q(y,z))) /\
   ((exists x y. Q(x,y)) ==> (forall x. R(x,x))) ==>
   (forall x. exists y. R(x,y))>>;;

trans "p_thirtyseven" p37

let p39 =
 <<~(exists x. forall y. P(y,x) <=> ~P(y,y))>>;;

trans "p_thirtynine" p39

let p40 =
 <<(exists y. forall x. P(x,y) <=> P(x,x))
  ==> ~(forall x. exists y. forall z. P(z,y) <=> ~P(z,x))>>;;

trans "p_fourty" p40

let p41 =
 <<(forall z. exists y. forall x. P(x,y) <=> P(x,z) /\ ~P(x,x))
  ==> ~(exists z. forall x. P(x,z))>>;;

trans "p_fourtyone" p41


let p44 =
 <<(forall x. P(x) ==> (exists y. G(y) /\ H(x,y)) /\
   (exists y. G(y) /\ ~H(x,y))) /\
   (exists x. J(x) /\ (forall y. G(y) ==> H(x,y))) ==>
   (exists x. J(x) /\ ~P(x))>>;;

trans "p_fourtyfour" p44


(* ------------------------------------------------------------------------- *)
(* Example from Manthey and Bry, CADE-9.                                     *)
(* ------------------------------------------------------------------------- *)

let p55 =
 <<lives(agatha) /\ lives(butler) /\ lives(charles) /\
   (killed(agatha,agatha) \/ killed(butler,agatha) \/
    killed(charles,agatha)) /\
   (forall x y. killed(x,y) ==> hates(x,y) /\ ~richer(x,y)) /\
   (forall x. hates(agatha,x) ==> ~hates(charles,x)) /\
   (hates(agatha,agatha) /\ hates(agatha,charles)) /\
   (forall x. lives(x) /\ ~richer(x,agatha) ==> hates(butler,x)) /\
   (forall x. hates(agatha,x) ==> hates(butler,x)) /\
   (forall x. ~hates(x,agatha) \/ ~hates(x,butler) \/ ~hates(x,charles))
   ==> killed(agatha,agatha) /\
       ~killed(butler,agatha) /\
       ~killed(charles,agatha)>>;;

trans "p_fiftyfive" p55

let p57 =
 <<P(f((a),b),f(b,c)) /\
   P(f(b,c),f(a,c)) /\
   (forall (x) y z. P(x,y) /\ P(y,z) ==> P(x,z))
   ==> P(f(a,b),f(a,c))>>;;

trans "p_fiftyseven" p57

(* ------------------------------------------------------------------------- *)
(* See info-hol, circa 1500.                                                 *)
(* ------------------------------------------------------------------------- *)

let p58 =
 <<forall P Q R. forall x. exists v. exists w. forall y. forall z.
    ((P(x) /\ Q(y)) ==> ((P(v) \/ R(w))  /\ (R(z) ==> Q(v))))>>;;

trans "p_fiftyeight" p58

let p59 =
 <<(forall x. P(x) <=> ~P(f(x))) ==> (exists x. P(x) /\ ~P(f(x)))>>;;

trans "p_fiftynine" p59

let p60 =
 <<forall x. P(x,f(x)) <=>
            exists y. (forall z. P(z,y) ==> P(z,f(x))) /\ P(x,y)>>;;

trans "p_sixty" p60

(* ------------------------------------------------------------------------- *)
(* From Gilmore's classic paper.                                             *)
(* ------------------------------------------------------------------------- *)

let gilmore_1 =
 <<exists x. forall y z.
      ((F(y) ==> G(y)) <=> F(x)) /\
      ((F(y) ==> H(y)) <=> G(x)) /\
      (((F(y) ==> G(y)) ==> H(y)) <=> H(x))
      ==> F(z) /\ G(z) /\ H(z)>>;;

trans "gilmore_one" gilmore_1


let gilmore_3 =
 <<exists x. forall y z.
        ((F(y,z) ==> (G(y) ==> H(x))) ==> F(x,x)) /\
        ((F(z,x) ==> G(x)) ==> H(z)) /\
        F(x,y)
        ==> F(z,z)>>;;

trans "gilmore_three" gilmore_3

let gilmore_4 =
 <<exists x y. forall z.
        (F(x,y) ==> F(y,z) /\ F(z,z)) /\
        (F(x,y) /\ G(x,y) ==> G(x,z) /\ G(z,z))>>;;

trans "gilmore_four" gilmore_4

let gilmore_5 =
 <<(forall x. exists y. F(x,y) \/ F(y,x)) /\
   (forall x y. F(y,x) ==> F(y,y))
   ==> exists z. F(z,z)>>;;

trans "gilmore_five" gilmore_5

let gilmore_6 =
 <<forall x. exists y.
        (exists u. forall v. F(u,x) ==> G(v,u) /\ G(u,x))
        ==> (exists u. forall v. F(u,y) ==> G(v,u) /\ G(u,y)) \/
            (forall u v. exists w. G(v,u) \/ H(w,y,u) ==> G(u,w))>>;;

trans "gilmore_six" gilmore_6

let gilmore_7 =
 <<(forall x. K(x) ==> exists y. L(y) /\ (F(x,y) ==> G(x,y))) /\
   (exists z. K(z) /\ forall u. L(u) ==> F(z,u))
   ==> exists v w. K(v) /\ L(w) /\ G(v,w)>>;;

trans "gilmore_seven" gilmore_7

let gilmore_8 =
 <<exists x. forall y z.
        ((F(y,z) ==> (G(y) ==> (forall u. exists v. H(u,v,x)))) ==> F(x,x)) /\
        ((F(z,x) ==> G(x)) ==> (forall u. exists v. H(u,v,z))) /\
        F(x,y)
        ==> F(z,z)>>;;

trans "gilmore_eight" gilmore_8;;

let steamroller = <<((forall x. P1(x) ==> P0(x)) /\ (exists x. P1(x))) /\
   ((forall x. P2(x) ==> P0(x)) /\ (exists x. P2(x))) /\
   ((forall x. P3(x) ==> P0(x)) /\ (exists x. P3(x))) /\
   ((forall x. P4(x) ==> P0(x)) /\ (exists x. P4(x))) /\
   ((forall x. P5(x) ==> P0(x)) /\ (exists x. P5(x))) /\
   ((exists x. Q1(x)) /\ (forall x. Q1(x) ==> Q0(x))) /\
   (forall x. P0(x)
              ==> (forall y. Q0(y) ==> R(x,y)) \/
                  ((forall y. P0(y) /\ S0(y,x) /\
                              (exists z. Q0(z) /\ R(y,z))
                              ==> R(x,y)))) /\
   (forall x y. P3(y) /\ (P5(x) \/ P4(x)) ==> S0(x,y)) /\
   (forall x y. P3(x) /\ P2(y) ==> S0(x,y)) /\
   (forall x y. P2(x) /\ P1(y) ==> S0(x,y)) /\
   (forall x y. P1(x) /\ (P2(y) \/ Q1(y)) ==> ~(R(x,y))) /\
   (forall x y. P3(x) /\ P4(y) ==> R(x,y)) /\
   (forall x y. P3(x) /\ P5(y) ==> ~(R(x,y))) /\
   (forall x. (P4(x) \/ P5(x)) ==> exists y. Q0(y) /\ R(x,y))
   ==> exists x y. P0(x) /\ P0(y) /\
                   exists z. Q1(z) /\ R(y,z) /\ R(x,y)>>;;

(* trans "steamroller" steamroller;; *)
